package substitution;

import java.util.HashMap;

import org.svvrl.goal.core.Message;
import org.svvrl.goal.core.aut.AlphabetType;
import org.svvrl.goal.core.aut.BuchiAcc;
import org.svvrl.goal.core.aut.OmegaUtil;
import org.svvrl.goal.core.aut.State;
import org.svvrl.goal.core.aut.Transition;
import org.svvrl.goal.core.aut.fsa.FSA;
import org.svvrl.goal.core.aut.fsa.NBWCreator;
import org.svvrl.goal.core.logic.Literal;
import org.svvrl.goal.core.logic.Proposition;
import org.svvrl.goal.gui.action.ExecutionException;

public class SubstitutionConstruction {

	private FSA pattern =null;
	private FSA scope = null;
	private int substitutionSstate;
	
	public SubstitutionConstruction(FSA scp, FSA pat, int subId) {
		super();
		
		if (!OmegaUtil.isNBW(pat))
			throw new IllegalArgumentException(
					Message.onlyForFSA(BuchiAcc.class));
		
		this.pattern = pat;
		this.scope = scp;
		this.substitutionSstate = subId;
	}
	
	public FSA substitute() throws ExecutionException {
		FSA result = new NBWCreator().create();
		
		// pattern states
		HashMap<State, State> transfoPattern = new HashMap<State, State>(); 
		for (State s : pattern.getStates()){
			State new_s = result.createState();
			transfoPattern.put(s,new_s);
		}
		
		// scope states
		HashMap<State, State> transfoScope = new HashMap<State, State>(); 
		for (State s : scope.getStates())
			if (s.getID() != substitutionSstate){
				State new_s = result.createState();
				transfoScope.put(s,new_s);
			}
		
		// pattern & scope alphabet
		result.expandAlphabet(pattern.getAtomicPropositions());
		result.expandAlphabet(scope.getAtomicPropositions());
		
		// get restrictions from outgoing transitions
		// TODO : many atomic propositions over an outgoing transition from the substitution state!!
		String restriction = "";
		for (Transition t_sub_out : scope.getTransitionsFromState(scope.getStateByID(substitutionSstate)))
			if (AlphabetType.PROPOSITIONAL.getLiterals(t_sub_out.getLabel()).size() > 1)
				throw new ExecutionException("TODO : many atomic propositions over an outgoing transition from the substitution state!!");
			else 
				restriction += " " + new Literal(new Proposition(t_sub_out.getLabel()), true);
				
		// pattern transitions with restriction
		for (Transition t_pat : pattern.getTransitions()){			
			result.createTransition(
					transfoPattern.get(t_pat.getFromState()), 
					transfoPattern.get(t_pat.getToState()), 
					AlphabetType.PROPOSITIONAL.formatLabel(t_pat.getLabel() + restriction));
		}
		
		// scope transitions
		for (Transition t_scope : scope.getTransitions()){			
			// transitions not concerning substitution state
			if (t_scope.getFromState().getID() != substitutionSstate && t_scope.getToState().getID() != substitutionSstate)
				result.createTransition(
						transfoScope.get(t_scope.getFromState()), 
						transfoScope.get(t_scope.getToState()), 
						t_scope.getLabel());
			
			// ongoing transitions to substitution state (non strict)
			if (t_scope.getFromState().getID() != substitutionSstate && t_scope.getToState().getID() == substitutionSstate)
				for (Transition tp : pattern.getTransitionsFromState(pattern.getInitialState())){
					result.createTransition(
							transfoScope.get(t_scope.getFromState()), 
							transfoPattern.get(tp.getToState()), 
							AlphabetType.PROPOSITIONAL.formatLabel(t_scope.getLabel() + " " + tp.getLabel()));
			}
			
			// outgoing transitions from substitution state (strict)
			if (t_scope.getFromState().getID() == substitutionSstate && t_scope.getToState().getID() != substitutionSstate){
				for(State s : ((BuchiAcc) pattern.getAcc()).get())
					result.createTransition(
							transfoPattern.get(s), 
							transfoScope.get(t_scope.getToState()), 
							t_scope.getLabel());
			}
			
			// TODO : cycling transitions over substitution state!!
			if (t_scope.getFromState().getID() == substitutionSstate && t_scope.getToState().getID() == substitutionSstate)
				throw new ExecutionException("TODO : cycling transitions over substitution state!!");
		}
		
		// initial state
		if (scope.getInitialState().getID() == substitutionSstate)
			result.setInitialState(transfoPattern.get(pattern.getInitialState()));
		else result.setInitialState(transfoScope.get(scope.getInitialState()));
		
		// final states
		for (State s : scope.getStates())
			if (s.getID() != substitutionSstate && ((BuchiAcc) scope.getAcc()).contains(s)) 
				((BuchiAcc) result.getAcc()).add(transfoScope.get(s));
		
		if (((BuchiAcc) scope.getAcc()).contains(scope.getStateByID(substitutionSstate)))
			for (State s : pattern.getStates())
				if (((BuchiAcc) pattern.getAcc()).contains(s)) 
					((BuchiAcc) result.getAcc()).add(transfoPattern.get(s));
		
		return result;
	}

}
